//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by NeoLogic are Copyright (C) 2001 NeoLogic Systems, Inc.
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.dataMapper;

import java.lang.String;
import java.io.IOException;
import org.jdom.JDOMException;

/**
 *
 * This interface provides support for language-independent <B>data binding</B>.
 * A <CODE>DataMapper</CODE> instance is obtained by using the 
 * <CODE>createDataMapper()</CODE> (and companions) static functions of 
 * <CODE>DataMapperFactory</CODE>. The <CODE>marshall()</CODE> and 
 * <CODE>marshallToURL()</CODE> functions map the given object to a serializable 
 * representation that is either returned as a <CODE>String</CODE> to the caller 
 * or written to the location passed to the factory when this data mapper 
 * instance was obtained. <CODE>unmarshall()</CODE> instanciates an object or 
 * objects based on the serialized representation. Both marshalling and 
 * unmarshalling may utilize optional schema data given to the factory when this 
 * data mapper instance was created.
 */
 
public interface DataMapper
{
	/**
	 *
	 * Creates a serialized <CODE>String</CODE> representation of the given object(s).
	 * The marshalling process may use the optional schema document given to the
	 * constructor of this data mapper instance.
	 *
	 * @param object - Object to be serialized.
	 * @return String - Serialized <CODE>String</CODE> representation of the given object(s).
	 * @throws Exception -
	 *    if the value of object is <CODE>null</CODE>.
	 *    if the location of the schema documents is invalid, malformed or inaccessible.
	 */
	public String marshall(Object object) throws Exception;

	/**
	 *
	 * Writes a serialized representation of the given object(s) to the instance document
	 * location given to the constructor. The marshalling process may use the optional
	 * XML schema document given to the constructor of this data mapper instance.
	 *
	 * @param object - Object to be serialized.
	 * @throws Exception -
	 *   if the value of object is null.
	 *   if the location of either the instance or schema documents are invalid,
	 *       malformed or inaccessible.
	 */
	public void marshallToURL(Object object) throws Exception;

	 /**
	  *
	  * Creates an object or tree of objects based on the contents of the XML
	  * instance and optional schema documents whose location were passed to the
	  * constructor of this data mapper. The unmarshalling process may use the optional
	  * XML schema document given to the constructor.
	  *
	  * @return Object - Object or tree of related objects instanciated based on
	  * the contents of the XML instance and optional schema documents this data
	  * mapper refers to.
	  * @throws Exception - if the location of either the instance or schema documents are invalid,
	  *    malformed or inaccessible.
	  */
	public Object unmarshall() throws Exception;
}

