//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by NeoLogic are Copyright (C) 2001 NeoLogic Systems, Inc.
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.dataMapper;

import java.lang.String;
import java.io.*;
import java.net.*;
import java.util.*;
import fipaos.util.DIAGNOSTICS;
import fipaos.agent.dataMapper.*;

/**
 *
 * This is a factory class used to create the proper type of <CODE>DataMapper</CODE>
 * instance given the location of an instance document. (<CODE>DataMapper</CODE>
 * objects are used to support language-independent <B>data binding</B>.) When a
 * <CODE>DataMapper</CODE> class is loaded it registers itself by passing a
 * <CODE>DataMapperFactory.Creator</CODE> instance to the <CODE>register()</CODE>
 * function of <CODE>DataMapperFactory</CODE>. This results in a binding between
 * the location of instance documents passed to the
 * <CODE>createDataMapperByURL()</CODE>, <CODE>createDataMapperByFile()</CODE> and
 * <CODE>createDataMapper()</CODE> functions of this class and the realizations of
 * the objects these documents describe. The
 * <CODE>dataMapperFactory.properties</CODE> properties file is used to insure that
 * all standard <CODE>DataMapper</CODE> subclasses are loaded before they're needed
 * by application code. The processes of marshalling and unmarshalling object
 * instances may utilize optional schema data given to the factory when a data
 * mapper instance was created.
 *
 * Note: This class has neither instance variables not instance methods. As such,
 * it should not be instanciated.
 */
 
public class DataMapperFactory
{
	/**
	 * List of registered creators. Each subclass of DataMapper should
	 * register a creator. The getInstance function of the factory iterates
	 * through this list in reverse-chronological order until a creator
	 * returns a DataMapper that can support the given URL.
	 */
	private static List __creators = new LinkedList();
	
	/**
	 *
	 * This is a <B>functoid</B> class. Every active subclass of
	 * <CODE>DataMapper</CODE> passes an instance of this class to the
	 * <CODE>register()</CODE> function of <CODE>DataMapperFactory</CODE>. The
	 * <CODE>createIfSupported()</CODE> function of
	 * <CODE>DataMapperFactory.Creator</CODE> takes URL <CODE>String</CODE>s that refer
	 * to an instance document and optional schema document. Each
	 * <CODE>DataMapper</CODE> subclass will implement this function to return a
	 * <CODE>DataMapper</CODE> instance for those instance documents that it supports.
	 */
	public static interface Creator
	{
		/**
		 *
		 * This function takes URL <CODE>String</CODE>s that refer to an instance document
		 * and optional schema document. Each <CODE>DataMapper</CODE> subclass will
		 * implement this function to return a <CODE>DataMapper</CODE> instance for those
		 * instance documents that it supports.
		 *
		 * @param data_url - A URL <CODE>String</CODE> that refers to the instance document this data
		 *   mapper is bound to.
		 * @param schema_url - An optional URL <CODE>String</CODE> that refers to an schema that this data
		 *   mapper may use in the marshalling and unmarshalling process.
		 * @return <CODE>DataMapper</CODE> that supports the given instance document, or
		 *    <CODE>null</CODE> if the document is not supported.
		 */
		public DataMapper createIfSupported(String data_url, String schema_url);
	}
	
	static
	{
		try
		{
			// Get any dataMapperFactory.properties
			ResourceBundle bundle = ResourceBundle.getBundle( "fipaos.agent.dataMapper.dataMapperFactory" );
			if( bundle != null )
			{
				// Enumerate through properties
				Enumeration enum = bundle.getKeys();
				while( enum.hasMoreElements() )
				{
					// Get key (interface name)
					String key = (String) enum.nextElement();
					
					// Get value (instance name)
					String value = bundle.getString( key );
					
					try
					{
						// Force loading of class defs (and hence static initialisers)
						Class.forName( value );
						Class.forName( key );
					}
					catch( Throwable t )
					{
						DIAGNOSTICS.println( "Some error whilst initialising dataMapper classes", t );
					}
				}
			}
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Some error whilst process dataMapper properties file", t );
		}
	}
	
	/**
	 *
	 * When a <CODE>DataMapper</CODE> class is loaded it registers itself by passing a
	 * <CODE>DataMapperFactory.Creator</CODE> instance to this function. This results
	 * in a binding between the location of instance documents passed to the
	 * <CODE>createDataMapperByURL()</CODE>, c<CODE>reateDataMapperByFile()</CODE> and
	 * <CODE>createDataMapper()</CODE> functions and the realizations of the objects
	 * these documents describe.
	 *
	 * @param creator - A <CODE>DataMapperFactory.Creator</CODE> having a
	 * <CODE>createIfSupported()</CODE> function that takes a URL <CODE>String</CODE>s
	 * that refer to an instance document and optional schema document. Each
	 * <CODE>DataMapper</CODE> subclass will implement this function to return a
	 * <CODE>DataMapper</CODE> instance for those instance documents that it supports.
	 */
	public static void register(DataMapperFactory.Creator creator)
	{
		__creators.add( creator );
	}
	
	/**
	 *
	 * Creates a <CODE>DataMapper</CODE> subclass based on the type of instance
	 * document referred to by the <CODE>data_url</CODE> parameter, or <CODE>null</CODE>
	 * if none of the registered data mapper classes are capable of support the given
	 * instance document type.
	 *
	 * @param data_url - A URL <CODE>String</CODE> that refers to the instance document a data mapper
	 *  is to be bound to.
	 * @param schema_url - An optional URL <CODE>String</CODE> that refers to an schema that the data
	 *  mapper may use in the marshalling and unmarshalling process.
	 * @return <CODE>DataMapper</CODE> that supports the given instance document, or
	 *    <CODE>null</CODE> if the document is not supported.
	 */
	public static DataMapper createDataMapperByURL(String data_url, String schema_url)
	{
		DataMapper data_mapper = null;
		Iterator iterator = __creators.iterator();
		
		while( iterator.hasNext() )
		{
			DataMapperFactory.Creator creator = (DataMapperFactory.Creator) iterator.next();
			
			data_mapper = creator.createIfSupported( data_url, schema_url );
			if( data_mapper != null )
				break;
		}
		
		return data_mapper;
	}
	
	/**
	 *
	 * Creates a <CODE>DataMapper</CODE> subclass based on the type of instance
	 * document referred to by the <CODE>data_url</CODE> parameter, or null if none of
	 * the registered data mapper classes are capable of support the given instance
	 * document type.
	 *
	 * @param data_file - A file path that refers to the instance document a data mapper is to be bound to.
	 * @param schema_url - An optional URL <CODE>String</CODE> that refers to an schema that the data
	 * mapper may use in the marshalling and unmarshalling process.
	 * @return <CODE>DataMapper</CODE> that supports the given instance document, or
	 *   <CODE>null</CODE> if the document is not supported.
	 */
	public static DataMapper createDataMapperByFile(String data_file, String schema_url)
	{
		File file = new File( data_file );

		return createDataMapper( file, schema_url );
	}

	/**
	 *
	 * Creates a <CODE>DataMapper</CODE> subclass based on the type of instance
	 * document referred to by the <CODE>data_url</CODE> parameter, or
	 * <CODE>null</CODE> if none of the registered data mapper classes are capable of
	 * supporting the given instance document type.
	 *
	 * @param data_url - A URL that refers to the instance document a data mapper is to be bound to.
	 * @param schema_url - An optional URL <CODE>String</CODE> that refers to an schema that the data
	 * mapper may use in the marshalling and unmarshalling process.
	 * @return <CODE>DataMapper</CODE> that supports the given instance document, or
	 *    <CODE>null</CODE> if the document is not supported.
	 */
	public static DataMapper createDataMapper(URL data_url, String schema_url)
	{
		return createDataMapperByURL( data_url.toExternalForm(), schema_url );
	}

	/**
	 *
	 * Creates a <CODE>DataMapper</CODE> subclass based on the type of instance
	 * document referred to by the <CODE>data_url</CODE> parameter, or
	 * <CODE>null</CODE> if none of the registered data mapper classes are capable of
	 * supporting the given instance document type.
	 *
	 * @param data_file - A <CODE>File</CODE> that refers to the instance document a data mapper is to
	 *  be bound to.
	 * @param schema_url - An optional URL <CODE>String</CODE> that refers to an schema that the data
	 *  mapper may use in the marshalling and unmarshalling process.
	 * @return <CODE>DataMapper</CODE> that supports the given instance document, or
	 *   <CODE>null</CODE> if the document is not supported.
	 */
	public static DataMapper createDataMapper(File data_file, String schema_url)
	{
		DataMapper dataMapper;
		
		try
		{
			// Attempt to create a URL from the given File
			URL url = data_file.toURL();

			dataMapper = createDataMapper( url, schema_url );
		}
		catch(Throwable f)
		{
			dataMapper = null;
		}
		
		return dataMapper;
	}
}

