//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by NeoLogic are Copyright (C) 2001 NeoLogic Systems, Inc.
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.dataMapper;

import java.lang.String;
import java.util.Map;
import java.util.Hashtable;
import java.io.IOException;
import java.io.File;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import fipaos.agent.profile.RuntimeProfile;
import fipaos.agent.profile.RuntimeProfileImpl;
import fipaos.agent.dataMapper.DataMapper;
import fipaos.agent.dataMapper.DataMapperFactory;
import fipaos.tool.configurator.ScriptVars;
import fipaos.tool.configurator.ScriptEditor;
import fipaos.tool.configurator.TwoWayMapping;
import fipaos.util.DocumentRetriever;

/**
 *
 * This class provides support for shell script and batch file data binding.
 * When the ScriptDataMapper class
 * is loaded it registers itself with the DataMapperFactory. This results in a
 * binding between the location of shell script or batch file passed to the
 * DataMapperFactory and the realizations of the objects that document
 * describes. A ScriptLDataMapper instance is obtained by passing the location of a
 * shell script and batch file to one of the  createDataMapper() (and companions) static
 * functions of DataMapperFactory. The marshall() and marshallToURL() functions
 * map the given object to a serializable representation that is either returned
 * as a String to the caller or written to the shell script or batch file location
 * that was passed to the factory
 * when this data mapper instance was obtained. unmarshall() instanciates
 * an object or objects based on the serialized representation. The implemention
 * of this data mapper does not support schemas. As such, the marshalling
 * and unmarshalling processes will ignore an optional schema given to the
 * factory when this data mapper instance was created.
 *
 */
 
public class ScriptDataMapper implements DataMapper
{
	private String _data_url;
	private ScriptEditor _script_editor;

	private static TwoWayMapping __jit_mappings;
	private static TwoWayMapping __dbg_lvl_mappings;
	
	static {
		DataMapperFactory.register( new DataMapperFactory.Creator() {
			public DataMapper createIfSupported(String data_url, String schema_url) {
				if( data_url.toLowerCase().indexOf( "setupfipaos" ) != -1 )
					return new ScriptDataMapper( data_url );
				else
					return null;
			}
		}
		);

			// set up the jit options TwoWayMapping
		__jit_mappings = new TwoWayMapping();
		
		__jit_mappings.put(ScriptVars.JIT_NORMAL, ScriptVars.JIT_VAL1);
		__jit_mappings.put(ScriptVars.JIT_NOJIT, ScriptVars.JIT_VAL2);
		__jit_mappings.put(ScriptVars.JIT_NOHOTSPOT, ScriptVars.JIT_VAL3);
		
		// set up the debug level TwoWayMapping
		__dbg_lvl_mappings = new TwoWayMapping();
		
		__dbg_lvl_mappings.put(new Integer(1), ScriptVars.DBG_LVL_VAL1);
		__dbg_lvl_mappings.put(new Integer(2), ScriptVars.DBG_LVL_VAL2);
		__dbg_lvl_mappings.put(new Integer(3), ScriptVars.DBG_LVL_VAL3);
		__dbg_lvl_mappings.put(new Integer(4), ScriptVars.DBG_LVL_VAL4);
		__dbg_lvl_mappings.put(new Integer(5), ScriptVars.DBG_LVL_VAL5);
	}
	
/**
 *
 * Initializes a newly created ScriptDataMapper object using the given shell
 * script or batch document.
 *
 * @param data_url - A URL string that refers to the script or batch file
 * this data mapper is bound to.
 *
 */
	public ScriptDataMapper(String data_url)
	{
		_data_url = data_url;
	}
	
/**
 *
 * Creates a serialized String representation of the given object(s). This String
 * should contain a set of valid shell script or batch file commands for this platform.
 *
 * @param object - Object to be serialized.
 *
 * @return String - A String that contains a set of valid shell script or batch
 * file commands for this platform.
 *
 * @throws Exception -
 * if the value of object is null or invalid.
 *
 */
	public String marshall(Object object) throws Exception
	{
		if( _script_editor == null )
			setScriptEditor( new ScriptEditor( new URL( _data_url ), ScriptVars.SETUPFIPAOS_VARS ) );

		updateScriptEditor( (RuntimeProfile) object );
		
		return _script_editor.getScript();
	}
	
/**
 *
 * Writes a serialized representation of the given object(s) to the shell script
 * or batch file given to the constructor.
 *
 * @param object - Object to be serialized.
 *
 * @throws Exception -
 * if the value of object is null.
 * if the location of either the shell script or batch file is invalid, malformed or inaccessible.
 *
 */
	public void marshallToURL(Object object) throws Exception
	{
		String text = marshall( object );
		
		// Attempt to create a URL from _url
		java.net.URL url = new java.net.URL( _data_url );
		
		// Attempt to open a connection to the URL
		URLConnection connection = url.openConnection();
		
		OutputStream stream = connection.getOutputStream();
		
		stream.write(text.getBytes(), 0, text.length());
		
		stream.flush();
	}
	
/**
 *
 * Creates an object or tree of objects based on the contents of the shell script 
 * or batch file that was passed to the constructor of this data mapper.
 *
 * @return Object - Object or tree of releated objects instanciated based on
 * the contents of the environment variables set in the shell script or batch file this data 
 * mapper refers to.
 *
 * @throws Exception -
 * if the location of either the shell script or batch file is invalid or inaccessible.
 *
 */
	public Object unmarshall() throws Exception
	{
		RuntimeProfileImpl runtime_profile = new RuntimeProfileImpl();
		
		try
		{
			setScriptEditor( new ScriptEditor( new URL( _data_url ), ScriptVars.SETUPFIPAOS_VARS ) );
			loadProfile( runtime_profile );
		}
		catch (Exception e)
		{
			System.out.println(e);
			System.out.println("Could not unmarshall " + _data_url );
			System.exit(-1);
		}

		return runtime_profile;
	}
	
/**
 *
 * Writes the script represented by the given RuntimeProfile to the location
 * given to the constructor of this data mapper.
 *
 * @param runtime_profile - RuntimeProfile object with the environment variable
 * settings to be written to the script or batch file.
 *
 * @throws Exception -
 * if the value of runtime_profile is null.
 * if the location of either the shell script or batch file is invalid or inaccessible.
 *
 */
	public void saveProfile(RuntimeProfile runtime_profile) {
		updateScriptEditor( runtime_profile );
		_script_editor.saveNewConfiguration();
	} // end saveProfile()
	
	protected ScriptEditor getScriptEditor() {
		return _script_editor;
	} // end getScriptEditor()
	
	protected void setScriptEditor(ScriptEditor script_editor) {
		_script_editor = script_editor;
	} // setScriptEditor()
	
	private void loadProfile( RuntimeProfile runtime_profile ) {
		
		Map script_vars = _script_editor.getScriptVariables();
		
		runtime_profile.setFipaosPath((String)script_vars.get(ScriptVars.SETUPFIPAOS_FIPAOS));
		runtime_profile.setFipaosVer((String)script_vars.get(ScriptVars.SETUPFIPAOS_FIPAOS_VER));
		runtime_profile.setJdkPath((String)script_vars.get(ScriptVars.SETUPFIPAOS_JDK));
		runtime_profile.setPlatformPath((String)script_vars.get(ScriptVars.SETUPFIPAOS_PLATFORM));
		
		// select debug combo box option
		String debug_select = (String)script_vars.get(ScriptVars.SETUPFIPAOS_DEBUGVER);
		
		runtime_profile.setDebugger(debug_select != null && debug_select.equals(ScriptVars.DEBUG_VAL1));
		
		// select loader gui combo box option
		String loader_select = (String)script_vars.get(ScriptVars.SETUPFIPAOS_LOADER_GUI);
		
		runtime_profile.setHasGui(loader_select != null && loader_select.equals(ScriptVars.LOADER_GUI_VAL1));
		
		String key;
		Integer value;
		
		// select jit combo box option
		key = (String)script_vars.get(ScriptVars.SETUPFIPAOS_NOJIT);
		
		if (key == null) {
			runtime_profile.setNoJit(ScriptVars.JIT_NORMAL);
		}
		else {
			value = (Integer)__jit_mappings.get(key);
			runtime_profile.setNoJit(value);
		}
		
		// select memory debug level combo box option
		key = (String)script_vars.get(ScriptVars.SETUPFIPAOS_DBG_LVL_MEM);
		
		if (key == null) {
			runtime_profile.setDebugMemLvl(ScriptVars.DEBUG_LVL_DEFAULT);
		}
		else {
			value = (Integer)__dbg_lvl_mappings.get(key);
			runtime_profile.setDebugMemLvl(value);
		}
		
		// select file debug level combo box option
		key = (String)script_vars.get(ScriptVars.SETUPFIPAOS_DBG_LVL_FILE);
		
		if (key == null) {
			runtime_profile.setDebugFileLvl(ScriptVars.DEBUG_LVL_DEFAULT);
		}
		else {
			value = (Integer)__dbg_lvl_mappings.get(key);
			runtime_profile.setDebugFileLvl(value);
		}
		
	} // end loadProfile()
	
	private void updateScriptEditor(RuntimeProfile runtime_profile) {
		
		Hashtable return_script_values = new Hashtable();
		
		return_script_values.put(ScriptVars.SETUPFIPAOS_FIPAOS, runtime_profile.getFipaosPath());
		return_script_values.put(ScriptVars.SETUPFIPAOS_JDK, runtime_profile.getJdkPath());
		return_script_values.put(ScriptVars.SETUPFIPAOS_FIPAOS_VER, runtime_profile.getFipaosVer());
		return_script_values.put(ScriptVars.SETUPFIPAOS_PLATFORM, runtime_profile.getPlatformPath());
		
		if (runtime_profile.getDebugger() == true) {
			return_script_values.put(ScriptVars.SETUPFIPAOS_DEBUGVER, ScriptVars.DEBUG_VAL1);
		}
		else {
			return_script_values.put(ScriptVars.SETUPFIPAOS_DEBUGVER, ScriptVars.DEBUG_VAL2);
		}
		return_script_values.put(ScriptVars.SETUPFIPAOS_NOJIT, __jit_mappings.get(runtime_profile.getNoJit()));
		if (runtime_profile.getHasGui() == true) {
			return_script_values.put(ScriptVars.SETUPFIPAOS_DEBUGVER, ScriptVars.LOADER_GUI_VAL1);
		}
		else {
			return_script_values.put(ScriptVars.SETUPFIPAOS_DEBUGVER, ScriptVars.LOADER_GUI_VAL2);
		}
		
		return_script_values.put(ScriptVars.SETUPFIPAOS_DBG_LVL_MEM, __dbg_lvl_mappings.get(runtime_profile.getDebugMemLvl()));
		return_script_values.put(ScriptVars.SETUPFIPAOS_DBG_LVL_FILE, __dbg_lvl_mappings.get(runtime_profile.getDebugFileLvl()));
		
		_script_editor.setScriptVariables(return_script_values);
		
	} // end updateScriptEditor()
	
}

