//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by NeoLogic are Copyright (C) 2001 NeoLogic Systems, Inc.
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.dataMapper;

import java.lang.String;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URLConnection;
import fipaos.agent.dataMapper.DataMapper;
import fipaos.agent.dataMapper.DataMapperFactory;
import fipaos.util.DocumentRetriever;
import fipaos.util.XMLDataBinding;

/**
 *
 * This class provides support for XML data binding. When the
 * <CODE>XMLDataMapper</CODE> class is loaded it registers itself with the
 * <CODE>DataMapperFactory</CODE>. This results in a binding between the location
 * of XML instance documents passed to the <CODE>DataMapperFactory</CODE> and the
 * realizations of the objects these documents describe. An
 * <CODE>XMLDataMapper</CODE> instance is obtained by passing the location of an
 * XML instance document to one of the  <CODE>createDataMapper()</CODE> (and
 * companions) static functions of <CODE>DataMapperFactory</CODE>. The
 * <CODE>marshall()</CODE> and <CODE>marshallToURL()</CODE> functions map the given
 * object to a serializable representation that is either returned as a
 * <CODE>String</CODE> to the caller or written to the XML instance document
 * location that was passed to the factory when this data mapper instance was
 * obtained. <CODE>unmarshall()</CODE> instanciates an object or objects based on
 * the serialized representation. Both marshalling and unmarshalling may utilize
 * optional schema data given to the factory when this data mapper instance was created.
 */
 
public class XMLDataMapper implements DataMapper
{
	private String _data_url;
	private String _schema_url;
	
	static {
		DataMapperFactory.register( new DataMapperFactory.Creator() {
			public DataMapper createIfSupported(String data_url, String schema_url) {
				if( data_url.endsWith( ".profile" ) )
					return new XMLDataMapper( data_url, schema_url );
				else
					return null;
			}
		}
		);
	}
	
	/** Initializes a newly created <CODE>XMLDataMapper</CODE> object using the given
	 * XML instance document and optional schema document.
	 *
	 * @param data_url - A URL string that refers to the XML instance document
	 * this data mapper is bound to.
	 * @param schema_url - An optional URL stirng that refers to an XML schema
	 * that this data mapper may use in the marshalling and unmarshalling process.
	 */
	public XMLDataMapper(String data_url, String schema_url)
	{
		_data_url = data_url;
		_schema_url = schema_url;
	}
	
	/**
	 *
	 * Creates a serialized <CODE>String</CODE> representation of the given object(s).
	 * The marshalling process may use the optional XML schema document given to the
	 * constructor of this data mapper instance.
	 *
	 * @param object - Object to be serialized.
	 * @return String - Serialized <CODE>String</CODE> representation of the given object(s).
	 * @throws Exception -
	 *    if the value of <CODE>Object</CODE> is <CODE>null</CODE>.
	 *    if the location of the schema documents is invalid, malformed or inaccessible.
	 */
	public String marshall(Object object) throws Exception
	{
		return XMLDataBinding.marshall( object );
	}
	
	/** Writes a serialized representation of the given object(s) to the instance
	 * document location given to the constructor. The marshalling process may use the
	 * optional XML schema document given to the constructor of this data mapper
	 * instance.
	 *
	 * @param object - Object to be serialized.
	 * @throws Exception -
	 *    if the value of <CODE>Object</CODE> is <CODE>null</CODE>.
	 *    if the location of either the instance or schema documents are invalid,
	 *        malformed or inaccessible.
	 */
	public void marshallToURL(Object object) throws Exception
	{
		String xml = marshall( object );
		
		// Attempt to create a URL from _url
		java.net.URL url = new java.net.URL( _data_url );
		
		// Attempt to open a connection to the URL
		URLConnection connection = url.openConnection();
		
		OutputStream stream = connection.getOutputStream();
		
		stream.write(xml.getBytes(), 0, xml.length());
		
		stream.flush();
	}
	
	/** Creates an object or tree of objects based on the contents of the XML
	 * instance and optional schema documents whose location were passed to the
	 * constructor of this data mapper. The unmarshalling process may use the optional
	 * XML schema document given to the constructor.
	 *
	 * @return Object - Object or tree of releated objects instanciated based on the contents
	 * of the XML instance and optional schema documents this data mapper refers to.
	 * @throws Exception -
	 *    if the location of either the instance or schema documents are invalid,
	 *        malformed or inaccessible.
	 */
	public Object unmarshall() throws Exception
	{
		String document = DocumentRetriever.getDocFromURL( _data_url );
		
		if( document == null )
			return null;
		else
			return XMLDataBinding.unmarshall( document );
	}
	
}

