//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

package fipaos.mts.service;

//----------------------------------------------------------------------

import fipaos.agent.profile.*;
import fipaos.agent.dataMapper.*;
import fipaos.mts.*;
import fipaos.mts.service.preparse.*;
import fipaos.skill.db.*;
import fipaos.util.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;

import java.util.*;
import java.io.*;

//----------------------------------------------------------------------

/**
 * This is an implementation of the ACC as described within the FIPA99 specification
 * It acts as a gateway between the internal transports and the external transports,
 * allowing messages to be exchanged between various FIPA Agent platforms
 * <br>
 * <br>
 */
public class AgentCommunicationChannel
{
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	/**
	 * Name of the ACC at the transport level
	 */
	public static String ACC_NAME = "acc";
	
	/**
	 * Name of the ACC's database of remote platforms
	 */
	public static String DATABASE_NAME = "acc-db";
	
	/**
	 * Indicates if ACC should propogate updated MTP addresses
	 */
	public static final boolean PROPOGATE_ADDRS = System.getProperty( "acc.propogate.addrs", "false" ).equals( "true" );
	
	/**
	 * Indicates if FIPA-OS bootstrap file format should be used
	 */
	public static final boolean FIPAOS_BOOTSTRAP = System.getProperty( "acc.bootstrap.fipaosformat", "false" ).equals( "true" );
	
	/**
	 * Indicates if ACC should attempt to retrieve a bootstrapping file containing
	 * MTP addresses (this is legacy behaviour)
	 */
	public static final boolean BOOTSTRAP_FILES = System.getProperty( "acc.bootstrap.usefile", "false" ).equals( "true" );
	
	//----------------------------------------------------------------------
	// STATIC VARIABLES
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * Reference to "internal" stack
	 */
	private PreParserService _internal_stack;
	
	/**
	 * Reference to the "external" stack
	 */
	private PreParserService _external_stack;
	
	/**
	 * List of internal/external address for this
	 */
	private List _addresses = new LinkedList();
	
	/**
	 * HAP address for this ACC
	 */
	private String _hap;
	
	/**
	 * Database containing addresses for known platforms
	 */
	private Database _platforms;
	
	/**
	 * Count of the number of update() threads active within the ACC
	 */
	private int _update_threads = 0;
	
	/**
	 * ACCProfile containing configuration details for this
	 */
	private ACCProfile _acc_profile;
	
	/**
	 * PlatformProfile containing configuration details for the platform this is
	 * part of
	 */
	private PlatformProfile _platform_profile;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR(S)
	//----------------------------------------------------------------------
	
	/**
	 * Creates an ACC using the given profile for initialisation parameters
	 */
	public AgentCommunicationChannel( PlatformProfile platform_profile, ACCProfile acc_profile )
	throws InitialisationException
	{
		// Save HAP address
		_hap = platform_profile.getHAPName();
		
		// Save profile details
		_acc_profile = acc_profile;
		_platform_profile = platform_profile;
		
		
		DIAGNOSTICS.println( "Setting up comms" );
		
		// Setup transports & stacks
		setupTransports( acc_profile );
		
		
		DIAGNOSTICS.println( "Publishing details" );
		
		// Publish our address details
		publishAddresses( acc_profile );
		
		
		DIAGNOSTICS.println( "Retreiving remote platform details" );
		
		// Deal with collecting details about remote platforms
		dealWithRemotePlatforms( acc_profile );
	}
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------
	
	private void dealWithRemotePlatforms( ACCProfile acc_profile )
	throws InitialisationException
	{
		// Get the profile of the database to use
		DatabaseProfile db_profile = acc_profile.getDatabaseProfile();
		
		try
		{
			// Open database to store remote platforms' addresses
			_platforms = DatabaseFactory.getDatabase( _hap + "-" + DATABASE_NAME, db_profile );
		}
		catch( Throwable t )
		{
		}
		
		// Check the Database object was initialised OK
		if( _platforms == null )
		{
			throw new InitialisationException( "Couldn't create database for some reason" +
			" - profile = " +  db_profile );
		}
		
		// Create/open database
		try
		{
			_platforms.openDatabase();
		}
		catch( DatabaseException e )
		{
			try
			{
				_platforms.createDatabase();
			}
			catch( DatabaseException f )
			{
				DIAGNOSTICS.println( "Couldn't open OR create database for remote ACC addresses", f,
				DIAGNOSTICS.LEVEL_MAX );
			}
		}
		
		try
		{
			// Now attempt to update the database with information from the ACC profile
			_platforms.startTransaction( false );
			
			// For each of the platforms specified in the ACC profile, update the details
			// within our database
			List remote_aps = acc_profile.getRemoteAgentPlatformProfiles();
			DIAGNOSTICS.println( "Remote-APS = " + remote_aps );
			
			// Check that we actually have some remote-aps specified
			if( remote_aps != null )
			{
				Iterator i = remote_aps.iterator();
				while( i.hasNext() )
				{
					// Get next RemoteAgentPlatformProfile object
					RemoteAgentPlatformProfile ap_profile = (RemoteAgentPlatformProfile) i.next();
					
					// Check if an associated version is within the database
					HAPAddresses hap_addrs = (HAPAddresses) _platforms.findObject( ap_profile.getHAPName() );
					if( hap_addrs != null )
					{
						
						// Update the database entry
						hap_addrs._addresses_location = ap_profile.getAddressesLocation();
						hap_addrs._hap_name = ap_profile.getHAPName();
						
						// Update the database
						_platforms.updateObject( hap_addrs );
						
					}
					else // No previous HAP addresses for this remote platform
					{
						
						// Create a HAPAddresses object to contain information about the
						// remote platform
						hap_addrs = new HAPAddresses( ap_profile.getHAPName(), null,
						ap_profile.getAddressesLocation() );
						
						// Add it to the database
						_platforms.addObject( hap_addrs );
						
					}
				}
			}
			
			// Now attempt to pull the addresses from the web-addresses given
			Enumeration enum = _platforms.getIDs();
			while( enum.hasMoreElements() )
			{
				// Get next HAP address
				String hap = (String) enum.nextElement();
				
				// Get associated HAPAddresses object
				HAPAddresses hap_addrs = (HAPAddresses) _platforms.findObject( hap );
				
				// Update addresses it contains + propogate our addresses to these ACC's
				update( hap_addrs );
			}
			
			// End database transaction
			_platforms.endTransaction();
			
			
			// Wait for all updates to complete
			waitForUpdates();
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when getting remote platforms' addressses: ", t, DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	
	/**
	 * Initialises the transports used by this ACC given an ACCProfile
	 */
	private void setupTransports( ACCProfile acc_profile ) throws InitialisationException
	{
		// Create CMS
		CommMultiplexService cms = new CommMultiplexService();
		CommMultiplexService cms2 = new CommMultiplexService();
		String acc_name = ACC_NAME + '@' + _hap;
		
		// Attempt to create internal transports
		Map internal_transports = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		Iterator i = acc_profile.getInternalAddresss().iterator();
		
		InternalAddress int_addr;
		
		while( i.hasNext() )
		{
			int_addr = (InternalAddress)i.next();
			
			// Get URL for NS of next transport
			URL ns = new URL(int_addr.getAddress());
			
			// Attempt to instantiate transport
			InternalMTP internal = TransportProvider.getDefaultInternalMTP( acc_name, ns );
			
			
			
			// Check if succesful
			if( internal != null )
			{
				DIAGNOSTICS.println("Found InternalMTP NS: " + ns, this, DIAGNOSTICS.LEVEL_2);
				
				// Add address to our list of addresses
				_addresses.addAll( internal.getAddresses() );
				
				// Initialise transport with regard to CMS
				internal.initialise( cms, null );
				
				// Iterate through protocols and add the MTP against it
				List protocols = internal.getProtocols();
				Iterator j = protocols.iterator();
				while( j.hasNext() )
				{
					// Add to our Map of internal transports
					internal_transports.put( j.next(), internal );
				}
			}
			else
			{
				// Display warning message
				DIAGNOSTICS.println( "Error starting MTP using NS-URL: " + ns, DIAGNOSTICS.LEVEL_3 );
			}
		}
		
		// Attempt to create external transports
		Map external_transports = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		i = acc_profile.getExternalAddresss().iterator();
		
		ExternalAddress ext_addr;
		
		while( i.hasNext() )
		{
			ext_addr = (ExternalAddress)i.next();
			
			// Get URL for NS of next transport
			URL ns = new URL(ext_addr.getAddress());
			
			// Attempt to instantiate transport
			ExternalMTP external = TransportProvider.getDefaultExternalMTP( acc_name, ns );
			
			// Check if succesful
			if( external != null )
			{
				DIAGNOSTICS.println("Found ExternalMTP NS: " + ns, this, DIAGNOSTICS.LEVEL_2);
				
				// Add address to our list of addresses
				_addresses.addAll( external.getAddresses() );
				
				// Initialise transport with regard to CMS
				external.initialise( cms2, null );
				
				// Iterate through protocols and add the MTP against it
				List protocols = external.getProtocols();
				Iterator j = protocols.iterator();
				while( j.hasNext() )
				{
					// Add to our Map of internal transports
					external_transports.put( j.next(), external );
				}
			}
			else
			{
				// Display warning message
				DIAGNOSTICS.println( "Error starting MTP using NS-URL: " + ns, DIAGNOSTICS.LEVEL_3 );
			}
		}
		
		// Create internal stack
		_internal_stack = new InternalStack();
		ServiceStack stack = new ServiceStack();
		stack.add( _internal_stack );
		stack.add( new BufferService() );
		cms.initialise( internal_transports );
		stack.add( cms );
		stack.initialise();
		
		// Create external stack
		_external_stack = new ExternalStack();
		stack = new ServiceStack();
		stack.add( _external_stack );
		stack.add( new BufferService() );
		stack.add( new ParserService() );
		cms2.initialise( external_transports );
		stack.add( cms2 );
		stack.initialise();
	}
	
	
	/**
	 * Publish addresses of this ACC to the location given in the ACCProfile provided
	 */
	private void publishAddresses( ACCProfile acc_profile )
	{
		// Create the addresses document
		String document = deparseAddresses( _hap, _addresses );
		
		// Write it to the specified location
		String location = acc_profile.getLocalAddressesLocation();
		
		try
		{
			// Attempt to create a file in the given location
			FileOutputStream fos = new FileOutputStream( location );
			fos.write( document.getBytes() );
			fos.flush();
			fos.close();
		}
		catch( IOException e )
		{
			DIAGNOSTICS.println( "An error occured whilst publishing the ACC's addresses: ",
			e, this, DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	
	/**
	 * Deals with incoming messages from the internal stack
	 */
	private void handleInternal( Message msg )
	{
		DIAGNOSTICS.println( "ACC dealing with message from internal transport" );
		
		// Check if messages are returned outgoing messages
		if( msg.getEnvelope().getErrorCode() != Service.NONE )
		{
			handleInternalError( msg );
			return;
		}
		
		// Strip any of our addresses from the AID
		AgentID aid = (AgentID) msg.getEnvelope().getIntendedReceiver().get(0);
		
		DIAGNOSTICS.println("Internal msg. receiver =  " + aid, this, DIAGNOSTICS.LEVEL_2);
		
		removeHAPAddresses( aid );
		
		DIAGNOSTICS.println("Receiver without HAP  " + aid, this, DIAGNOSTICS.LEVEL_2);
		
		// Check if this is an internal or externally bound message
		String hap = getHAP( aid );
		if( hap.equalsIgnoreCase( _hap ) )
		{
			// Check if the message is for the ACC!
			if( aid.getName().startsWith( "acc@" ) )
			{
				handleForACC( msg );
			}
			else
			{
				// Internally bound messages
				_internal_stack.outgoing( msg );
			}
		}
		else
		{
			// Add addresses to intended-receiver AID's for destination platforms, if known
			addHAPAddresses( (AgentID) msg.getEnvelope().getIntendedReceiver().get(0) );
			
			// TODO: Generalise this!!!!
			// Updates AMA objects with external MTP addresses
			Object content = msg.getACL().getContentObject();
			if( content != null )
			{
				if( content instanceof AgentManagementAction )
				{
					updateMessageContent( (AgentManagementAction) content );
				}
				else if( content instanceof AgentManagementResult )
				{
					updateMessageContent( (AgentManagementResult) content );
				}
			}
			
			// Add local addresses to sender AID's if message is from this platform
			addLAPAddresses( msg );
			
			// Forward message to external transports
			_external_stack.outgoing( msg );
		}
	}
	
	
	/**
	 * Deals with incoming messages from the external stack
	 */
	private void handleExternal( Message msg )
	{
		DIAGNOSTICS.println( "ACC dealing with message from external transport" );
		
		// Check if messages are returned outgoing messages
		if( msg.getEnvelope().getErrorCode() != Service.NONE )
		{
			handleExternalError( msg );
			return;
		}
		
		
		// Add our mark to the envelope!
		msg.getEnvelope().newSubEnvelope();
		msg.getEnvelope().setReceived( new Received( null, (URL) _addresses.get( 0 ),
		null, null, new UTCTime() ) );
		
		
		// Check if we're dealing with a multi-recipient message
		if( msg.getEnvelope().getIntendedReceiver().size() > 1 )
		{
			// Split apart message
			Iterator i = msg.getEnvelope().getIntendedReceiver().iterator();
			while( i.hasNext() )
			{
				// Get next aid
				AgentID aid = (AgentID) i.next();
				
				try
				{
					// Duplicate the message, & resend
					
					// Clone Message
					Message msg_clone = (Message) msg.clone();
					
					// Update intended reciever field
					List new_recv = new LinkedList();
					new_recv.add( aid );
					msg_clone.getEnvelope().setIntendedReceiver( new_recv );
					
					// Re-send message
					handleExternal( msg_clone );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Unable to split multi-recipient message", t, this );
				}
			}
			
			return;
		}
		
		// Strip any of our addresses from the AID
		AgentID aid = (AgentID) msg.getEnvelope().getIntendedReceiver().get(0);
		
		DIAGNOSTICS.println("External msg. Receiver = " + aid, this, DIAGNOSTICS.LEVEL_2);
		
		removeHAPAddresses( aid );
		
		DIAGNOSTICS.println("Receiver without HAP  " + aid, this, DIAGNOSTICS.LEVEL_2);
		
		// Check if this is an internal or externally bound message
		String hap = getHAP( aid );
		if( hap.equalsIgnoreCase( _hap ) )
		{
			
			DIAGNOSTICS.println("Msg internally bound", this, DIAGNOSTICS.LEVEL_2);
			
			// Check if the message is for the ACC!
			if( aid.getName().startsWith( "acc@" ) )
			{
				handleForACC( msg );
			}
			else
			{
				// Internally bound messages
				_internal_stack.outgoing( msg );
			}
		}
		else
		{
			DIAGNOSTICS.println("Msg externally bound", this, DIAGNOSTICS.LEVEL_2);
			
			// Add addresses to intended-receiver AID's for destination platforms, if known
			addHAPAddresses( (AgentID) msg.getEnvelope().getIntendedReceiver().get(0) );
			
			// Forward message to external transports
			_external_stack.outgoing( msg );
		}
	}
	
	
	/**
	 * Handle messages bound for this!
	 */
	private void handleForACC( Message msg )
	{
		DIAGNOSTICS.println( "ACC dealing with message sent to it directly" );
		
		try
		{
			
			// Default case - this is an update pertaining to an ACC's addresses
			String content = (String) msg.getACL().getContentObject();
			HAPAddresses hap_addrs = parseAddresses( content );
			
			DIAGNOSTICS.println( "Received updated address information from " +
			hap_addrs._hap_name, DIAGNOSTICS.LEVEL_MAX );
			
			// Add back into database
			_platforms.startTransaction( false );
			
			// Check if we got addresses OK
			if( hap_addrs != null )
			{
				_platforms.addObject( hap_addrs );
				DIAGNOSTICS.println( "Succesfully updated address information in DB from " +
				hap_addrs._hap_name, DIAGNOSTICS.LEVEL_MAX );
			}
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when processing message for ACC: ", t, this,
			DIAGNOSTICS.LEVEL_3 );
		}
		finally
		{
			try
			{
				_platforms.endTransaction();
			}
			catch( Throwable t )
			{
			}
		}
	}
	
	
	/**
	 * Deals with outgoing internal messages which have not been sent for some reason
	 */
	private void handleInternalError( Message msg )
	{
		if( msg.getEnvelope().getSlotValue( Service.MSG_DELIVERY_FAILURE_REASON ) == null )
		{
			DIAGNOSTICS.println( "Message being sent internally failed, bouncing: " + msg );
			
			Envelope env = msg.getEnvelope();
			env.setIntendedReceiver( Arrays.asList( new Object[]
			{ env.getSender() } ) );
			env.setSlotValue( Service.MSG_DELIVERY_FAILURE_REASON, Service.ERROR_STRING[ env.getErrorCode() ] );
			env.setErrorCode( Service.NONE );
			
			handleInternal( msg );
		}
		else
		{
			DIAGNOSTICS.println( "Failure message being sent internally failed: " + msg );
		}
	}
	
	
	/**
	 * Deals with outgoing external messages which have not been sent for some reason
	 */
	private void handleExternalError( Message msg )
	{
		if( msg.getEnvelope().getSlotValue( Service.MSG_DELIVERY_FAILURE_REASON ) == null )
		{
			DIAGNOSTICS.println( "Message being sent externally failed, bouncing: " + msg );
			
			Envelope env = msg.getEnvelope();
			env.setIntendedReceiver( Arrays.asList( new Object[]
			{ env.getSender() } ) );
			env.setSlotValue( Service.MSG_DELIVERY_FAILURE_REASON, Service.ERROR_STRING[ env.getErrorCode() ] );
			env.setErrorCode( Service.NONE );
			
			DIAGNOSTICS.println( "Modified message: " + msg );
			
			handleExternal( msg );
		}
		else
		{
			DIAGNOSTICS.println( "Failure message being sent externally failed: " + msg );
		}
	}
	
	
	/**
	 * Gets the HAP of the given AgentID
	 */
	private String getHAP( AgentID aid )
	{
		// Get the name & pull out HAP name
		String name = aid.getName();
		int index = name.indexOf( '@' );
		
		// Check if HAP has been specified
		if( index != -1 )
		{
			// Extract HAP
			String hap = name.substring( index + 1 );
			return hap;
		}
		else
		{
			return "";
		}
	}
	
	/**
	 * Adds any known addresses of the LAP of the given messages sending AID
	 * if the sender is a member of the local HAP
	 */
	private void addLAPAddresses( Message msg )
	{
		Envelope env = msg.getEnvelope();
		AgentID aid1 = env.getSender();
		AgentID aid2 = msg.getACL().getSenderAID();
		
		// Get the HAP of the AgentID
		String hap = getHAP( aid1 );
		
		// Check if the message is from a local agent
		if( hap != null && hap.equals( _hap ) )
		{
			addURLs( aid1, _addresses );
			addURLs( aid2, _addresses );
		}
	}
	
	/**
	 * Adds any known addresses of the HAP of the given Agents' AID to the list of
	 * addresses within the AID, if not already present.
	 */
	private void addHAPAddresses( AgentID aid )
	{
		
		DIAGNOSTICS.println("Adding Addresses to AID : " + aid, this, DIAGNOSTICS.LEVEL_2);
		
		// Get the HAP of the AgentID
		String hap = getHAP( aid );
		
		DIAGNOSTICS.println("HAP for AID : " + hap, this, DIAGNOSTICS.LEVEL_2);
		
		// Only retrieve HAP addresses if not for the local platform
		if( !hap.equalsIgnoreCase( _hap ) )
		{
			HAPAddresses hap_addrs = null;
			
			try
			{
				_platforms.startTransaction( true );
				
				// Search database for HAP addresses for the given Agent's AID
				hap_addrs = (HAPAddresses) _platforms.findObject( hap );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error when accessing database: ", t, this, DIAGNOSTICS.LEVEL_MAX );
			}
			finally
			{
				try
				{
					_platforms.endTransaction();
				}
				catch( Throwable t )
				{
				}
			}
			
			// If we got HAPAddresses object, populate AID with them
			if( hap_addrs != null )
			{
				DIAGNOSTICS.println("Adding the following addresses" + hap_addrs._hap_addresses, this, DIAGNOSTICS.LEVEL_2);
				
				addURLs( aid, hap_addrs._hap_addresses );
			}
		}
	}
	
	
	/**
	 * Update the AgentManagementAction to contain the correct addresses
	 *
	 * TODO: Update to handle more cases
	 */
	private void updateMessageContent( AgentManagementAction ama )
	{
		// Check if it is a registration
		if( ama.getAction().equalsIgnoreCase( "register" ) )
		{
			// Check registration type
			if( ama.getDFAgentDescription() != null )
			{
				// Get AID
				AgentID aid = ama.getDFAgentDescription().getAgentID();
				
				// Udate AID as appropriate
				addURLs( aid, _addresses );
			}
			else if( ama.getAMSAgentDescription() != null ) 
			{
				// Get AID
				AgentID aid = ama.getAMSAgentDescription().getAgentID();
				
				// Udate AID as appropriate
				addURLs( aid, _addresses );
			}
		}
	}
	
	
	/**
	 * Update message content for AgentManagementResult objects
	 *
	 * TODO: Handle more cases
	 */
	private void updateMessageContent( AgentManagementResult amr )
	{
		// Get AMA type
		String action = amr.getAgentManagementAction().getAction();
		
		// Act appropriately
		if( action.equalsIgnoreCase( "get-description" ) )
		{
			// Update results
			APTDescription aptd = new APTDescription();
			
			// Create MTPDescription for external transports
			Map names = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
			Iterator i = _addresses.iterator();
			while( i.hasNext() )
			{
				try
				{
					// Convert to URL
					URL url = (URL) i.next();
					
					// Get profile name for MTP
					String profile = TransportProvider.getDefaultMTPProfileName( url.getProtocol() );
					
					// Is it valid?
					if( profile != null )
					{
						MTPDescription mtpd = null;
						
						// Create/retrieve profile
						if( names.containsKey( profile ) )
						{
							// Get MTPDescription already found
							mtpd = (MTPDescription) names.get( profile );
						}
						else
						{
							mtpd = new MTPDescription();
							mtpd.setMTPName( profile );
							aptd.addMTPD( mtpd );
							names.put( profile, mtpd );
						}
						
						// Add address
						mtpd.addAddress( url );
					}
				}
				catch( Throwable t )
				{
				}
			}
			
			// Add MTPD to results
			List l = new LinkedList( amr.getResultSet() );
			APDescription apd = (APDescription) ( l ).get( 0 );
			apd.setTransportDescription( aptd );
		}
	}
	
	
	/**
	 * Remove this HAP's addresses from the given AgentID
	 */
	private void removeHAPAddresses( AgentID aid )
	{
		// Get a list of the addresses associated with this AgentID
		List addrs = aid.getAddresses();
		
		// Remove the addresses for this
		Iterator i = _addresses.iterator();
		while( i.hasNext() )
		{
			URL url = (URL) i.next();
			addrs.remove( url );
		}
	}
	
	
	/**
	 * Generic method to add a List of URL's to the given AID
	 */
	private void addURLs( AgentID aid, List urls )
	{
		// Get addresses of target Agent
		List addresses = aid.getAddresses();
		
		// Iterate through platform addresses
		Iterator i = urls.iterator();
		while( i.hasNext() )
		{
			// Get next address
			URL address = (URL) i.next();
			
			// If address is not already spcified, add it to the list
			if( !addresses.contains( address ) )
			{
				addresses.add( ( (URL) address ).clone() );
			}
		}
	}
	
	
	/**
	 * Starts a new thread to deal with retreiving the addresses for an AP
	 */
	private void update( final HAPAddresses hap_addrs )
	{
		final Object parent = this;
		
		synchronized( parent )
		{
			_update_threads++;
		}
		
		// Create a new thread to go off and download the remote platforms addresses
		Thread t = new Thread( new Runnable()
		{
			public void run()
			{
				updateHAPDetails( parent, hap_addrs );
				if( hap_addrs._hap_addresses != null )
				{
					sendUpdateDetails( hap_addrs );
				}
			}
		}, "Remote Platform Address Retreiver Thread");
		t.start();
	}
	
	
	/**
	 * Send message to the ACC of the given HAP indicating our addresses
	 */
	private void sendUpdateDetails( HAPAddresses hap_addrs )
	{
		// Check if we should propogate addresses
		if( PROPOGATE_ADDRS )
		{
			
			// Send a message to the target ACC indicating our new addresses
			AgentID acc = new AgentID();
			acc.setName( "acc@" + hap_addrs._hap_name );
			acc.setAddresses( hap_addrs._hap_addresses );
			
			ACL acl = new ACL();
			acl.setPerformative( FIPACONSTANTS.INFORM );
			acl.setReceiverAID( acc );
			acl.setContentObject( deparseAddresses( _hap, _addresses ) );
			
			DIAGNOSTICS.println( "Attempting to send updated addresses to acc@" + hap_addrs._hap_name,
			this, DIAGNOSTICS.LEVEL_MAX );
			
			_external_stack.outgoing( new Message( acl ) );
		}
	}
	
	
	/**
	 * Send message to the ACC of the given HAP indicating our addresses
	 */
	private void updateHAPDetails( Object parent, HAPAddresses hap_addrs )
	{
		boolean treat_as_mtp = true;
		
		// Should we allow legacy bootstrapping?
		if( BOOTSTRAP_FILES )
		{
			// Retreive the address information
			String document = DocumentRetriever.getDoc( hap_addrs._addresses_location );

			// Check if we actually got anything!
			if( document != null )
			{
				// Attempt to parse the file
				parseAddresses( document, hap_addrs );

				// Display addresses found
				DIAGNOSTICS.println( "Found addresses " + hap_addrs._hap_addresses + " for platform " +
				hap_addrs._hap_name, this );

				// Indicate that URL isn't an MTP address
				treat_as_mtp = false;
				
				// Save object
				try
				{
					_platforms.startTransaction( false );

					_platforms.addObject( hap_addrs );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Error when updating DB: ", t, this, DIAGNOSTICS.LEVEL_MAX );
				}
				finally
				{
					try
					{
						_platforms.endTransaction();
					}
					catch( Throwable t )
					{
					}
				}
			}
		}
		
		
		// Treat URL as an MTP?
		if( treat_as_mtp )
		{
			try
			{
				// Attempt to convert location to URL
				URL url = new URL( hap_addrs._addresses_location );
				
				// URL must be transport address!
				hap_addrs._hap_addresses = new LinkedList();
				hap_addrs._hap_addresses.add( url );
				
				// Save object
				try
				{
					_platforms.startTransaction( false );
					
					_platforms.addObject( hap_addrs );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Error when updating DB: ", t, this, DIAGNOSTICS.LEVEL_MAX );
				}
				finally
				{
					try
					{
						_platforms.endTransaction();
					}
					catch( Throwable t )
					{
					}
				}
				
				DIAGNOSTICS.println( "Using \"" +
				hap_addrs._addresses_location + " as a MTP address for platform " +
				hap_addrs._hap_name, this );
			}
			catch( Throwable t )
			{
				// Display error message
				DIAGNOSTICS.println( "Couldn't retreive addresses for platform \"" +
				hap_addrs._hap_name + "\" from " +
				hap_addrs._addresses_location, DIAGNOSTICS.LEVEL_4 );
			}
		}
		
		// Indicate that we're done
		synchronized( parent )
		{
			_update_threads--;
			parent.notify();
		}
	}
	
	/**
	 * Waits for all "update" threads to complete
	 */
	private void waitForUpdates()
	{
		synchronized( this )
		{
			// Loop until number of update() threads reaches 0
			while( _update_threads > 0 )
			{
				try
				{
					// Wait for notify() to be invoked by an update() thread
					wait();
				}
				catch( InterruptedException e )
				{
				}
			}
		}
	}
	
	
	/**
	 * Update the addresses in the given HAPAddresses with those in the given String
	 */
	private static void parseAddresses( String document, HAPAddresses original )
	{
		HAPAddresses new_hap = parseAddresses( document );
		if( new_hap != null )
		{
			original._hap_addresses = new_hap._hap_addresses;
		}
	}
	
	
	/**
	 * Parse the given document into a list of addresses - returns null if the
	 * HAP details expected don't match those given
	 */
	private static HAPAddresses parseAddresses( String document )
	{
		List out = new LinkedList();
		HAPAddresses hap_addr = null;
		
		// Create a StringTokenizer to deal with the document
		StringTokenizer st = new StringTokenizer( document, ",\r\n", false );
		
		// Check the HAP name
		String hap = st.nextToken();
		if( hap != null )
		{
			// Check if file just contains IOR
			if( hap.startsWith( "IOR:" ) )
			{
				hap_addr = new HAPAddresses( null, out, null );
				out.add( new URL( hap ) );
			}
			else
			{
				hap_addr =  new HAPAddresses( hap, out, null );
			}
			
			// Iterate through the rest of the tokens
			while( st.hasMoreTokens() )
			{
				// Get the next token, and attempt to convert it into a URL
				try
				{
					String next = st.nextToken();
					
					if( !next.trim().equals( "" ) )
					{
						out.add( new URL( next ) );
					}
				}
				catch( Throwable t )
				{
				}
			}
		}
		
		return hap_addr;
	}
	
	/**
	 * Create a document containing a number of addresses through which an ACC can be
	 * contacted
	 */
	private static String deparseAddresses( String hap, List addrs )
	{
		StringBuffer out = new StringBuffer();
		
		// Construct the address document
		if( FIPAOS_BOOTSTRAP )
		{
			out.append( hap );
		}
		
		Iterator i = addrs.iterator();
		while( i.hasNext() )
		{
			URL next = (URL) i.next();
			
			if( FIPAOS_BOOTSTRAP )
			{
				if( next != null )
				{
					out.append( ',' );
					out.append( next.toString() );
				}
			}
			else if( next.getProtocol().equals( "IOR" ) )
			{
				out.append( next.toString() );
			}
		}
		
		return out.toString();
	}
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// INNER-CLASSES
	//----------------------------------------------------------------------
	
	/**
	 * Class to encapsulate data to be stored within a database
	 */
	private static class HAPAddresses implements DatabaseObject, Serializable
	{
		public String _addresses_location;
		public String _hap_name;
		public List _hap_addresses = new LinkedList();
		
		public HAPAddresses( String hap_name, List hap_addresses, String addresses_location )
		{
			_addresses_location = addresses_location;
			_hap_name = hap_name;
			_hap_addresses = hap_addresses;
		}
		
		public String getObjectID()
		{
			return _hap_name;
		}
	}
	
	/**
	 * Class to propogate incoming messages from the internal stack to the
	 * appropriate method in parent class
	 */
	private class InternalStack implements PreParserService
	{
		private PreParserService _next;
		
		/** This should be invoked when the service should be safely shutdown
		 */
		public void shutdown()
		{
			_next.shutdown();
		}
		
		/** Invoked when the undelying implementation should initialise itself.  Allows
		 * references to other PreParserService's before and after it in the stack
		 * to be passed
		 * @param previous Previous PreParserService instance (Incoming messages should be passed to this)
		 * @param next Next PreParserService instance (Outgoing messages should be passed to this)
		 */
		public void initialise(PreParserService previous,PreParserService next)
		{
			_next = next;
		}
		
		/** Outgoing messages are passed to <CODE>this</CODE> via this method
		 * @param msg Outgoing message
		 */
		public void outgoing(Message msg)
		{
			_next.outgoing( msg );
		}
		
		/** Incoming messages are passed to <CODE>this</CODE> via this method
		 * @param msg Incoming message
		 */
		public void incoming(Message msg)
		{
			handleInternal( msg );
		}
	}
	
	/**
	 * Class to propogate incoming messages from the external stack to the
	 * appropriate method in parent class
	 */
	private class ExternalStack implements PreParserService
	{
		private PreParserService _next;
		
		/** This should be invoked when the service should be safely shutdown
		 */
		public void shutdown()
		{
			_next.shutdown();
		}
		
		/** Invoked when the undelying implementation should initialise itself.  Allows
		 * references to other PreParserService's before and after it in the stack
		 * to be passed
		 * @param previous Previous PreParserService instance (Incoming messages should be passed to this)
		 * @param next Next PreParserService instance (Outgoing messages should be passed to this)
		 */
		public void initialise(PreParserService previous,PreParserService next)
		{
			_next = next;
		}
		
		/** Outgoing messages are passed to <CODE>this</CODE> via this method
		 * @param msg Outgoing message
		 */
		public void outgoing(Message msg)
		{
			_next.outgoing( msg );
		}
		
		/** Incoming messages are passed to <CODE>this</CODE> via this method
		 * @param msg Incoming message
		 */
		public void incoming(Message msg)
		{
			handleExternal( msg );
		}
	}
	
	//----------------------------------------------------------------------
	// MAIN/TESTING METHOD
	//----------------------------------------------------------------------
	
	public static void main( String args[] ) throws Throwable
	{
		// Attempt to load platform profile
		DataMapper dataMapper = DataMapperFactory.createDataMapperByFile( args[0], null );
		PlatformProfile platform_profile = (PlatformProfile) dataMapper.unmarshall();
		
		// Attempt to load ACC profile
		dataMapper = DataMapperFactory.createDataMapperByFile(
			new File( platform_profile.getProfileDirectory(), "acc.profile" ).getCanonicalPath(), null );
		ACCProfile acc_profile = (ACCProfile) dataMapper.unmarshall();
		
		// Create ACC
		new AgentCommunicationChannel( platform_profile, acc_profile );
	}
} // End class
