//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.tool.configurator;

import java.util.*;
import javax.swing.*;
import java.io.*;
import java.net.*;

public class ScriptEditor implements ScriptVars
{
	
	private static final String _separator = System.getProperty("file.separator");
	
	// store the script keys and values in arrays to preserve ordering
	private String _script_variables[];
	private String _script_values[];
	private String _script_lines[];
	
	// use a Map to convieniently pass the script data to the gui
	private Map _script_keys_values;
	
	// keep track of the line numbers that variables are stored on
	private Map _vars_to_lines;
	
	public static final String SETUP_UNIX		= "SetupFIPAOS";
	
	private boolean _is_windows;
	
	private String _url;
	private String _real_fipaos_location;
	private String _script;
	
	// LinkedList to allow quick access to editible vars
	private List _editable_vars;
	
	// Indicates if we're working in LEGACLY mode
	private boolean _legacy;
	
	/*********************************************************
	 *
	 * Constructor
	 *
	 *********************************************************/
	public ScriptEditor()
	{
		
		_legacy = true;
		
		_is_windows = isWindowsOS();
		
		// copy editable script var names into the linked list for quick access
		
		_editable_vars = new LinkedList();
		
		_editable_vars.addAll( SETUPFIPAOS_VARS );
		
	}
	
	/*********************************************************
	 *
	 * Constructor
	 *
	 * @param File. File of the script to be edited
	 *
	 * @param List. List of variables whose values can be set
	 *
	 *********************************************************/
	public ScriptEditor( File batch_file, List variables ) throws IOException
	{
		_is_windows = isWindowsOS();

		_editable_vars = new LinkedList( variables );		
		setFilename( batch_file.getCanonicalPath() );
		_script = FIPAOSConfiguratorTools.loadURLIntoString( _url );
		extractParameters();
	}
	
	
	/*********************************************************
	 *
	 * Constructor
	 *
	 * @param URL. URL of the script to be edited
	 *
	 * @param List. List of variables whose values can be set
	 *
	 *********************************************************/
	public ScriptEditor( URL url, List variables ) throws IOException
	{
		_is_windows = isWindowsOS();

		_editable_vars = new LinkedList( variables );		
		setURL( url.toExternalForm() );
		_script = FIPAOSConfiguratorTools.loadURLIntoString( _url );
		extractParameters();
	}
	
	
	/*********************************************************
	 *
	 * getURL()
	 *
	 * @return String. URL of the script being edited.
	 *
	 *********************************************************/
	public String getURL()
	{
		return _url;
	} // end getURL()
	
	/*********************************************************
	 *
	 * setURL()
	 *
	 * @param filename. URL of the script to be edited.
	 *
	 *********************************************************/
	public void setURL(String url)
	{
		_url = url;

		if ( _is_windows )
		{
			if( _url.endsWith( ".bat" ) == false )
				_url = _url + ".bat";
		}
		else
		{
			if( _url.endsWith( ".bat" ) )
				_url =
					_url.substring( 0, _url.length() - 4 );
		}
		_is_windows = _url.endsWith( ".bat" );
	} // end setURL()
	
	/*********************************************************
	 *
	 * setFilename()
	 *
	 * @param filename. Path of the script to be edited.
	 *
	 *********************************************************/
	public void setFilename(String filename)
	{
		try
		{
			setURL( new File( filename ).toURL().toExternalForm() );
		}
		catch (Throwable e)
		{
			System.out.println(e);
			System.out.println("ScriptEditor could not set filename; " + filename );
		}
	} // end setFilename()
	
	/*********************************************************
	 *
	 * isWindowsOS()
	 *
	 * @return boolean. True if this is a windows OS. False if not.
	 *
	 *********************************************************/
	public boolean isWindowsOS()
	{
	
		String os = System.getProperty("os.name");
		
		if (os.toLowerCase().indexOf("windows") == -1)
		{
		
			return false;
			
		}
		else
		{
		
			return true;
			
		}
		
	} // end isWindowsOS()
	
	/*********************************************************
	 *
	 * populate()
	 *
	 * @param installation. A string containing the location of the FIPA-OS
	 * installation that holds the script to be parse.
	 *
	 * @return boolean. True if the script variables were successfully parsed.
	 *
	 *********************************************************/
	public boolean populate(String installation)
	{
		
		_real_fipaos_location = installation;
		
		if (installation == null) // use JFileChooser to locate script
		{
			JFileChooser chooser = new JFileChooser(installation);
			
			chooser.setDialogTitle("Select location of setup script");
			
			int returnVal = chooser.showDialog(null, "Select");
			
			if(returnVal == JFileChooser.APPROVE_OPTION) {
				try {
					setFilename( chooser.getSelectedFile().getCanonicalPath() );
				}
				catch (Throwable t) {
					t.printStackTrace();
					
					return populate(null);
					
				} // end try
			}
		}
		else {
			setFilename( installation + _separator + "bat" + _separator + SETUP_UNIX );
		}
		
		try
		{
			_script = FIPAOSConfiguratorTools.loadFileIntoString(_url);
		}
		catch (Throwable t)
		{
			t.printStackTrace();
			
			return populate(null);
			
		} // end try
		
		extractParameters();

		return true;		
		
	} // end populate()
	
	/*********************************************************
	 *
	 * getValueOfScriptVariable()
	 *
	 * @param variable. The variable name to get the value for
	 *
	 * @return String. The value associated with the given variable name from the script. null if not found
	 *
	 *********************************************************/
	private String getValueOfScriptVariable(String variable)
	{
		
		for (int i = 0; i < _script_variables.length; i++)
		{
			
			// if this variable position contains the requires variable name
			if (_script_variables[i] != null && _script_variables[i].equals(variable))
			{
				
				return _script_values[i];
				
			} // end if
			
		} // end for
		
		// not found
		return null;
		
	} // end getValueOfScriptVariable()
	
	/*********************************************************
	 *
	 * setValueOfScriptVariable()
	 *
	 * @param variable. The variable name to set the value of
	 *
	 * @param value. The value to set the variable to
	 *
	 * @return boolean. True if the variable was set
	 *
	 *********************************************************/
	private boolean setValueOfScriptVariable(String variable, String value)
	{
		for (int i = 0; i < _script_variables.length; i++)
		{
			
			// if this variable position contains the requires variable name
			if (_script_variables[i] != null && _script_variables[i].equals(variable))
			{
				
				_script_values[i] = value;
				
				_script_keys_values.put(_script_variables[i], value);
				
				return true;
				
			} // end if
			
		} // end for
		
		// not found
		return false;
		
	} // end setValueOfScriptVariable()
	
	/*********************************************************
	 *
	 * extractParameters()
	 *
	 * Parses the script and extracts the parameters and values for later use.
	 *
	 *********************************************************/
	public void extractParameters()
	{
		
		StringTokenizer lines = new StringTokenizer(_script, "\n", false);
		
		String line;
		
		String prefix = "";
		
		if ( _is_windows )
		{
		
			prefix = "set ";
			
		}
		
		String name;
		String value;
		
		int set_present;
		int equals_present;
		
		int num_tokens = lines.countTokens();
		
		_vars_to_lines = new HashMap();
		
		_script_keys_values = new HashMap();
		
		_script_variables 	= new String[num_tokens];
		_script_values		= new String[num_tokens];
		_script_lines		= new String[num_tokens];
		
		int pos = 0;
		
		int line_number = 0;
		
		while (lines.hasMoreTokens())
		{
			
			line = lines.nextToken();
			
			_script_lines[line_number] = line;
			
			// if the line contains the set prefix and an = sign
			// then this is a parameter line (possibly rem'd out)
			set_present = line.toLowerCase().indexOf(prefix);
			equals_present = line.indexOf('=', set_present);
			
			if (set_present != -1 && equals_present != -1)
			{ // this is a parameter line
				
				name  = getName(line);
				value = getValue(line);
				
				// if variable name is not null
				if (name != null)
				{
					
					// if it is an editable variable
					if (_editable_vars == null || _editable_vars.contains(name))
					{

						_script_variables[pos] 	= name;
						_script_values[pos]	= value;
						
						_script_keys_values.put(name, value);
						
						_vars_to_lines.put(name, new Integer(line_number));
						
						pos++;
						
					} // end if editable
					
				} // end if valid variable
				
			} // end if variable line
			
			line_number++;
			
		} // end while
		
		// If a legacy client, then override the value of SETUPFIPAOS_FIPAOS with real location
		if ( _legacy )
			setValueOfScriptVariable( SETUPFIPAOS_FIPAOS, _real_fipaos_location );

	} // end extractParameters()
	
	/*********************************************************
	 *
	 * getScriptVariables()
	 *
	 * @return Map. The variables parsed from the script as key-value pairs
	 *
	 *********************************************************/
	public Map getScriptVariables()
	{
		
		return _script_keys_values;
		
	} // end getScriptVariables()
	
	/*********************************************************
	 *
	 * getName()
	 *
	 * @param line. The line in the script to examine.
	 *
	 * @return String. The variable name found on that line. null if none found.
	 *
	 *********************************************************/
	private String getName(String line)
	{
		
		int start;
		int end;
		
		String prefix = "";
		
		if (_is_windows == true)
		{
			
			prefix = "set ";
			
			start = line.toLowerCase().indexOf(prefix);
			
			end 	= line.indexOf('=', start);
			
		}
		else
		{ // no 'set' used to set unix environment vars
			
			start = 0;
			
			end 	= line.indexOf('=', start);
			
		}
		
		if (start != -1 && end != -1)
		{
			
			return line.substring(start + prefix.length(), end);
			
		}
		else
		{
			
			return null;
			
		}
		
	} // end getName()
	
	/*********************************************************
	 *
	 * getValue()
	 *
	 * @param line. The line in the script to examine.
	 *
	 * @return String. The value from the given line. null if none found.
	 *
	 *********************************************************/
	private String getValue(String line)
	{
		
		int start = line.indexOf('=');
		
		if (start != -1)
		{
			
			return line.substring(start + 1, line.length());
			
		}
		else
		{
			
			return null;
			
		}
		
	}	// end getValue()
	
	/*********************************************************
	 *
	 * setScriptVariables()
	 *
	 * @param variables. The values to write back to the script file
	 *
	 *********************************************************/
	public void setScriptVariables(Map variables)
	{
		
		Iterator keys = variables.keySet().iterator();
		
		String key;
		String value;
		Integer to_replace;
		int line_number;
		
		String new_line;
		
		// for each variable returned
		while (keys.hasNext())
		{
			
			// get the variable name
			key = (String)keys.next();
			
			// get the value
			value = (String)variables.get(key);
			
			// get the script line number to replace
			to_replace = (Integer)_vars_to_lines.get(key);
			
			// this can occur for variables that are used in the configurator but not in the GUI
			if (to_replace != null)
			{
				
				// convert to an int
				line_number = to_replace.intValue();
				
				if ( _is_windows == true )
				{
					
					new_line = "SET " + key + "=" + value;
					
				}
				else
				{
					
					new_line = key + "=" + value;
					
				}
				
				// replace the line in the script
				_script_lines[line_number] = new_line;
				
			}
			
		} // end while
		
		StringBuffer updated_script = new StringBuffer();
		
		String linebreak = System.getProperty("line.separator");
		
		for (int i = 0; i < _script_lines.length; i++)
		{
			
			updated_script.append(_script_lines[i]);
			updated_script.append(linebreak);
			
		}
		
		_script = updated_script.toString();
		
	} // end setScriptVariables()
	
	/*********************************************************
	 *
	 * getScript()
	 *
	 * @return String. The text of the script.
	 *
	 *********************************************************/
	public String getScript()
	{
		return _script;
	} // end getScript()
	
	/*********************************************************
	 *
	 * saveNewConfiguration()
	 *
	 * Save the updated information to the configuration files
	 *
	 *********************************************************/
	public void saveNewConfiguration()
	{
		
		// Save the script
		try
		{
			FIPAOSConfiguratorTools.saveStringIntoURL(_url, _script);
		}
		catch (Exception e)
		{
			System.out.println(e);
			System.out.println("Could not save " + _url.substring( _url.lastIndexOf( _separator ) ) );
		}
		
	} // end saveNewConfiguration()
	
} // end ScriptEditor.java
